from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any

from mipac.models.announcement import Announcement
from mipac.models.lite.role import PartialRole
from mipac.models.lite.user import BadgeRole, PartialUser
from mipac.models.note import Note
from mipac.types.follow import IFederationFollowCommon, IFederationFollower, IFederationFollowing
from mipac.types.meta import IPolicies
from mipac.types.page import IPage
from mipac.types.user import (
    EmailNotificationTypes,
    GetFrequentlyRepliedUsersResponse,
    IAchievement,
    IFfVisibility,
    IMeDetailedOnlySchema,
    IMeDetailedSchema,
    ITwoFactorBackupCodesStock,
    IUser,
    IUserDetailedNotMeOnlySchema,
    IUserDetailedNotMeSchema,
    IUserField,
    IUserList,
    IUserListMembership,
    IUserNotify,
    IUserRole,
    IUserSecurityKey,
    NotificationRecieveConfig,
    is_me_detailed,
    is_user_detailed_not_me,
)
from mipac.utils.format import str_to_datetime
from mipac.utils.util import DeprecatedClass

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager
    from mipac.manager.users.list import ClientUserListManager


__all__ = ("PartialUser", "Achievement", "MeDetailed")


class FollowCommon[FFC: IFederationFollowCommon]:
    def __init__(self, raw_follow: FFC, *, client: ClientManager) -> None:
        self._raw_follow: FFC = raw_follow
        self._client: ClientManager = client

    @property
    def id(self) -> str:
        return self._raw_follow["id"]

    @property
    def created_at(self) -> datetime:
        return str_to_datetime(self._raw_follow["created_at"])

    @property
    def follower_id(self) -> str:
        return self._raw_follow["follower_id"]

    @property
    def follower(self) -> UserDetailedNotMe | None:
        raw_follower = self._raw_follow.get("follower")
        return packed_user(raw_follower, client=self._client) if raw_follower else None

    @property
    def followee_id(self) -> str:
        return self._raw_follow["followee_id"]

    @property
    def followee(self) -> UserDetailedNotMe | None:
        raw_followee = self._raw_follow.get("followee")
        return packed_user(raw_followee, client=self._client) if raw_followee else None

    def _get(self, key: str) -> Any | None:
        return self._raw_follow.get(key)


class Follower(FollowCommon[IFederationFollower]):
    def __init__(self, raw_follow: IFederationFollower, *, client: ClientManager) -> None:
        super().__init__(raw_follow=raw_follow, client=client)


class Following(FollowCommon[IFederationFollowing]):
    def __init__(self, raw_following: IFederationFollowing, *, client: ClientManager) -> None:
        super().__init__(raw_follow=raw_following, client=client)


class Achievement:
    def __init__(self, detail: IAchievement):
        self.__detail: IAchievement = detail

    @property
    def name(self) -> str:
        return self.__detail["name"]

    @property
    def unlocked_at(self) -> int:
        return self.__detail["unlocked_at"]

    def _get(self, key: str) -> Any | None:
        return self.__detail.get(key)


class UserField:
    def __init__(self, raw_user_field: IUserField, *, client: ClientManager) -> None:
        self._raw_user_field: IUserField = raw_user_field
        self.__client: ClientManager = client

    @property
    def name(self) -> str:
        return self._raw_user_field["name"]

    @property
    def value(self) -> str:
        return self._raw_user_field["value"]

    def _get(self, key: str) -> Any | None:
        return self._raw_user_field.get(key)


@DeprecatedClass(remove_in_version="0.7.0")
class UserRole(BadgeRole[IUserRole]):
    def __init__(self, data: IUserRole, *, client: ClientManager) -> None:
        super().__init__(data, client=client)

    @property
    def id(self) -> str:
        return self._data["id"]

    @property
    def color(self) -> str | None:
        return self._data["color"]

    @property
    def description(self) -> str:
        return self._data["description"]

    @property
    def is_moderator(self) -> bool:
        return self._data["is_moderator"]

    @property
    def is_administrator(self) -> bool:
        return self._data["is_administrator"]


class UserDetailedNotMeOnly:
    def __init__(self, raw_user: IUserDetailedNotMeOnlySchema, *, client: ClientManager) -> None:
        self._raw_user: IUserDetailedNotMeOnlySchema = raw_user
        self._client: ClientManager = client

    @property
    def url(self) -> str | None:
        return self._raw_user["url"]

    @property
    def uri(self) -> str | None:
        return self._raw_user["uri"]

    @property
    def moved_to(self) -> str | None:
        return self._raw_user["moved_to"]

    @property
    def also_known_as(self) -> list[str] | None:
        return self._raw_user["also_known_as"]

    @property
    def created_at(self) -> datetime:
        return str_to_datetime(self._raw_user["created_at"])

    @property
    def updated_at(self) -> datetime | None:
        return (
            str_to_datetime(self._raw_user["updated_at"]) if self._raw_user["updated_at"] else None
        )

    @property
    def last_fetched_at(self) -> datetime | None:
        return (
            str_to_datetime(self._raw_user["last_fetched_at"])
            if self._raw_user["last_fetched_at"]
            else None
        )

    @property
    def banner_url(self) -> str | None:
        return self._raw_user["banner_url"]

    @property
    def banner_blurhash(self) -> str | None:
        return self._raw_user["banner_blurhash"]

    @property
    def is_locked(self) -> bool:
        return self._raw_user["is_locked"]

    @property
    def is_silenced(self) -> bool:
        return self._raw_user["is_silenced"]

    @property
    def is_suspended(self) -> bool:
        return self._raw_user["is_suspended"]

    @property
    def description(self) -> str | None:
        return self._raw_user["description"]

    @property
    def location(self) -> str | None:
        return self._raw_user["location"]

    @property
    def birthday(self) -> str | None:
        return self._raw_user["birthday"]

    @property
    def lang(self) -> str | None:
        return self._raw_user["lang"]

    @property
    def fields(self) -> list[UserField]:
        return [UserField(i, client=self._client) for i in self._raw_user["fields"]]

    @property
    def verified_links(self) -> list[str]:
        return self._raw_user["verified_links"]

    @property
    def followers_count(self) -> int:
        return self._raw_user["followers_count"]

    @property
    def following_count(self) -> int:
        return self._raw_user["following_count"]

    @property
    def notes_count(self) -> int:
        return self._raw_user["notes_count"]

    @property
    def pinned_note_ids(self) -> list[str]:
        return self._raw_user["pinned_note_ids"]

    @property
    def pinned_notes(self) -> list[Note]:
        return [Note(raw_note, client=self._client) for raw_note in self._raw_user["pinned_notes"]]

    @property
    def pinned_page_id(self) -> str | None:
        return self._raw_user["pinned_page_id"]

    @property
    def pinned_page(self) -> IPage | None:  # TODO: モデルに
        return self._raw_user["pinned_page"]

    @property
    def public_reactions(self) -> bool:
        return self._raw_user["public_reactions"]

    @property
    def following_visibility(self) -> IFfVisibility:
        return self._raw_user["following_visibility"]

    @property
    def followers_visibility(self) -> IFfVisibility:
        return self._raw_user["followers_visibility"]

    @property
    def two_factor_enabled(self) -> bool:
        return self._raw_user["two_factor_enabled"]

    @property
    def use_password_less_login(self) -> bool:
        return self._raw_user["use_password_less_login"]

    @property
    def security_keys(self) -> bool:
        return self._raw_user["security_keys"]

    @property
    def roles(self) -> list[PartialRole]:
        return [PartialRole(i, client=self._client) for i in self._raw_user["roles"]]

    @property
    def memo(self) -> str | None:
        return self._raw_user["memo"]

    @property
    def moderation_note(self) -> str | None:
        return self._raw_user.get("moderation_note")

    @property
    def is_following(self) -> bool | None:
        return self._raw_user.get("is_following")

    @property
    def is_followed(self) -> bool | None:
        return self._raw_user.get("is_followed")

    @property
    def has_pending_follow_request_from_you(self) -> bool | None:
        return self._raw_user.get("has_pending_follow_request_from_you")

    @property
    def has_pending_follow_request_to_you(self) -> bool | None:
        return self._raw_user.get("has_pending_follow_request_to_you")

    @property
    def is_blocking(self) -> bool | None:
        return self._raw_user.get("is_blocking")

    @property
    def is_blocked(self) -> bool | None:
        return self._raw_user.get("is_blocked")

    @property
    def is_muted(self) -> bool | None:
        return self._raw_user.get("is_muted")

    @property
    def is_renote_muted(self) -> bool | None:
        return self._raw_user.get("is_renote_muted")

    @property
    def notify(self) -> IUserNotify | None:
        return self._raw_user.get("notify")

    @property
    def with_replies(self) -> bool | None:
        return self._raw_user.get("with_replies")


class MeDetailedOnly:
    def __init__(self, raw_user: IMeDetailedOnlySchema, *, client: ClientManager) -> None:
        self._raw_user: IMeDetailedOnlySchema = raw_user
        self.__client: ClientManager = client

    @property
    def avatar_id(self) -> str | None:
        return self._raw_user["avatar_id"]

    @property
    def banner_id(self) -> str | None:
        return self._raw_user["banner_id"]

    @property
    def is_moderator(self) -> bool | None:
        return self._raw_user["is_moderator"]

    @property
    def is_admin(self) -> bool | None:
        return self._raw_user["is_admin"]

    @property
    def inject_featured_note(self) -> bool:
        return self._raw_user["inject_featured_note"]

    @property
    def receive_announcement_email(self) -> bool:
        return self._raw_user["receive_announcement_email"]

    @property
    def always_mark_nsfw(self) -> bool:
        return self._raw_user["always_mark_nsfw"]

    @property
    def auto_sensitive(self) -> bool:
        return self._raw_user["auto_sensitive"]

    @property
    def careful_bot(self) -> bool:
        return self._raw_user["careful_bot"]

    @property
    def auto_accept_followed(self) -> bool:
        return self._raw_user["auto_accept_followed"]

    @property
    def no_crawle(self) -> bool:
        return self._raw_user["no_crawle"]

    @property
    def prevent_ai_learning(self) -> bool:
        return self._raw_user["prevent_ai_learning"]

    @property
    def is_explorable(self) -> bool:
        return self._raw_user["is_explorable"]

    @property
    def is_deleted(self) -> bool:
        return self._raw_user["is_deleted"]

    @property
    def two_factor_backup_codes_stock(self) -> ITwoFactorBackupCodesStock:
        return self._raw_user["two_factor_backup_codes_stock"]

    @property
    def hide_online_status(self) -> bool:
        return self._raw_user["hide_online_status"]

    @property
    def has_unread_specified_notes(self) -> bool:
        return self._raw_user["has_unread_specified_notes"]

    @property
    def has_unread_mentions(self) -> bool:
        return self._raw_user["has_unread_mentions"]

    @property
    def has_unread_announcement(self) -> bool:
        return self._raw_user["has_unread_announcement"]

    @property
    def unread_announcements(self) -> list[Announcement]:
        return [
            Announcement(i, client=self.__client) for i in self._raw_user["unread_announcements"]
        ]

    @property
    def has_unread_antenna(self) -> bool:
        return self._raw_user["has_unread_antenna"]

    @property
    def has_unread_channel(self) -> bool:
        return self._raw_user["has_unread_channel"]

    @property
    def has_unread_notification(self) -> bool:
        return self._raw_user["has_unread_notification"]

    @property
    def has_pending_received_follow_request(self) -> bool:
        return self._raw_user["has_pending_received_follow_request"]

    @property
    def unread_notifications_count(self) -> int:
        return self._raw_user["unread_notifications_count"]

    @property
    def muted_words(self) -> list[list[str]]:
        return self._raw_user["muted_words"]

    @property
    def hard_muted_words(self) -> list[list[str]]:
        return self._raw_user["hard_muted_words"]

    @property
    def muted_instances(self) -> list[str]:
        return self._raw_user["muted_instances"]

    @property
    def notification_recieve_config(self) -> NotificationRecieveConfig:  # TODO: モデル化
        return self._raw_user["notification_recieve_config"]

    @property
    def email_notification_types(self) -> list[EmailNotificationTypes]:  # TODO: モデル化
        return self._raw_user["email_notification_types"]

    @property
    def achievements(self) -> list[Achievement]:
        return [Achievement(i) for i in self._raw_user["achievements"]]

    @property
    def logged_in_days(self) -> int:
        return self._raw_user["logged_in_days"]

    @property
    def policies(self) -> IPolicies:  # TODO: モデル化
        return self._raw_user["policies"]

    @property
    def email(self) -> str | None:
        return self._raw_user.get("email")

    @property
    def email_verified(self) -> bool | None:
        return self._raw_user.get("email_verified")

    @property
    def security_keys_list(self) -> list[IUserSecurityKey] | None:  # TODO: モデル化
        return self._raw_user.get("security_keys_list")


class UserDetailedNotMe[T: IUserDetailedNotMeSchema](PartialUser[T], UserDetailedNotMeOnly):  # pyright: ignore 直せない
    def __init__(self, raw_user: T, *, client: ClientManager) -> None:
        super().__init__(raw_user, client=client)


class MeDetailed(UserDetailedNotMe[IMeDetailedSchema], MeDetailedOnly):  # pyright: ignore 直せない
    def __init__(self, raw_user: IMeDetailedSchema, *, client: ClientManager):
        super().__init__(raw_user, client=client)


def packed_user(user: IUser, client: ClientManager) -> UserDetailedNotMe | MeDetailed:
    if is_user_detailed_not_me(user):
        return UserDetailedNotMe(user, client=client)
    if is_me_detailed(user):
        return MeDetailed(user, client=client)
    raise ValueError("Invalid user model")


class UserList:
    def __init__(self, raw_user_list: IUserList, *, client: ClientManager) -> None:
        self._raw_user_list: IUserList = raw_user_list
        self.__client: ClientManager = client

    @property
    def id(self) -> str:
        return self._raw_user_list["id"]

    @property
    def created_at(self) -> datetime:
        return str_to_datetime(self._raw_user_list["created_at"])

    @property
    def name(self) -> str:
        return self._raw_user_list["name"]

    @property
    def user_ids(self) -> list[str]:
        return self._raw_user_list["user_ids"]

    @property
    def is_public(self) -> bool:
        return self._raw_user_list["is_public"]

    @property
    def api(self) -> ClientUserListManager:
        return self.__client.user._create_client_user_list_manager(self.id)

    def _get(self, key: str) -> Any | None:
        return self._raw_user_list.get(key)


class UserListMembership:
    def __init__(
        self, raw_user_list_membership: IUserListMembership, *, client: ClientManager
    ) -> None:
        self.__raw_user_list_membership: IUserListMembership = raw_user_list_membership
        self.__client: ClientManager = client

    @property
    def id(self) -> str:
        return self.__raw_user_list_membership["id"]

    @property
    def created_at(self) -> datetime:
        return str_to_datetime(self.__raw_user_list_membership["created_at"])

    @property
    def user_id(self) -> str:
        return self.__raw_user_list_membership["user_id"]

    @property
    def user(self) -> PartialUser:
        return PartialUser(self.__raw_user_list_membership["user"], client=self.__client)

    @property
    def with_replies(self) -> bool:
        return self.__raw_user_list_membership["with_replies"]

    def _get(self, key: str) -> Any | None:
        return self.__raw_user_list_membership.get(key)


class FrequentlyRepliedUser:
    def __init__(
        self,
        raw_frequently_replied_user: GetFrequentlyRepliedUsersResponse,
        *,
        client: ClientManager,
    ):
        self._raw_frequently_replied_user: GetFrequentlyRepliedUsersResponse = (
            raw_frequently_replied_user
        )
        self.__client: ClientManager = client

    @property
    def user(self) -> UserDetailedNotMe | MeDetailed:
        return packed_user(self._raw_frequently_replied_user["user"], client=self.__client)

    @property
    def weight(self) -> int:
        return self._raw_frequently_replied_user["weight"]

    def _get(self, key: str) -> Any | None:
        return self._raw_frequently_replied_user.get(key)



from __future__ import annotations

from typing import TYPE_CHECKING, AsyncGenerator

from mipac.abstract.action import AbstractAction
from mipac.http import HTTPClient, Route
from mipac.models.note import Note
from mipac.models.roles import Role, RoleUser
from mipac.types.note import INote
from mipac.types.roles import IRole, IRoleUser
from mipac.utils.pagination import Pagination

if TYPE_CHECKING:
    from mipac.client import ClientManager


class RoleActions(AbstractAction):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        self.__session: HTTPClient = session
        self.__client: ClientManager = client

    async def get_list(self) -> list[Role]:
        """
        Get a list of roles from the API.
        Endpoint: `/api/roles/list`

        Returns
        -------
        list[Role]
            The role data.
        """
        res: list[IRole] = await self.__session.request(
            Route("POST", "/api/roles/list"), auth=True
        )

        return [Role(raw_role, client=self.__client) for raw_role in res]

    async def get(self, role_id: str) -> Role:
        """
        Get a role from the API.
        Endpoint: `/api/roles/show`

        Parameters
        ----------
        role_id : str
            The ID of the role to get.

        Returns
        -------
        Role
            The role data.
        """
        raw_role: IRole = await self.__session.request(
            Route("POST", "/api/roles/show"), auth=True, json={"roleId": role_id}
        )
        return Role(raw_role, client=self.__client)

    async def get_users(
        self,
        role_id: str,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        *,
        get_all: bool = False,
    ) -> AsyncGenerator[RoleUser, None]:
        """
        Get users in a role.
        Endpoint: `/api/roles/users`

        Parameters
        ----------
        role_id : str
            The ID of the role to get users.
        since_id : str, optional
            The ID of the user to get users after, by default None
        until_id : str, optional
            The ID of the user to get users before, by default None
        limit : int, optional
            The number of users to get, by default 10
        get_all : bool, optional
            Whether to get all users, by default False

        Yields
        ------
        AsyncGenerator[RoleUser, None]
            The role user data.
        """

        if limit > 100:
            raise ValueError("Limit cannot be greater than 100")

        if get_all:
            limit = 100

        body = {"roleId": role_id, "limit": limit, "sinceId": since_id, "untilId": until_id}

        pagination = Pagination[IRoleUser](
            self.__session, Route("POST", "/api/roles/users"), auth=True, json=body
        )
        while True:
            raw_users = await pagination.next()
            for raw_user in raw_users:
                yield RoleUser(raw_user, client=self.__client)

            if pagination.is_final or get_all is False:
                break

    async def get_notes(
        self,
        role_id: str,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        since_data: int | None = None,
        until_data: int | None = None,
        *,
        get_all: bool = False,
    ):
        """
        Get notes in a role.
        Endpoint: `/api/roles/notes`

        Parameters
        ----------
        role_id : str
            The ID of the role to get notes.
        limit : int, optional
            The number of notes to get, by default 10
        since_id : str, optional
            The ID of the note to get notes after, by default None
        until_id : str, optional
            The ID of the note to get notes before, by default None
        since_data : int, optional
            The timestamp of the note to get notes after, by default None
        until_data : int, optional
            The timestamp of the note to get notes before, by default None
        get_all : bool, optional
            Whether to get all notes, by default False

        Yields
        ------
        AsyncGenerator[Note, None]
            The note data.
        """

        if limit > 100:
            raise ValueError("Limit cannot be greater than 100")

        if get_all:
            limit = 100

        body = {
            "roleId": role_id,
            "limit": limit,
            "sinceId": since_id,
            "untilId": until_id,
            "sinceData": since_data,
            "untilData": until_data,
        }

        pagination = Pagination[INote](
            self.__session, Route("POST", "/api/roles/notes"), auth=True, json=body
        )
        while True:
            raw_notes = await pagination.next()
            for raw_note in raw_notes:
                yield Note(raw_note, client=self.__client)

            if pagination.is_final or get_all is False:
                break



import pathlib

from setuptools import setup

import versioneer

description = "A Python wrapper for the Misskey API"
readme_file = pathlib.Path(__file__).parent / "README.md"
with readme_file.open(encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as f:
    requirements = f.read().splitlines()

extras_require = {
    "dev": ["isort", "mypy", "pre-commit", "ruff"],
    "ci": ["mypy", "ruff"],
    "speed": ["orjson"],
    "doc": ["sphinx", "furo", "sphinxcontrib_trio", "sphinx-intl", "numpydoc"],
}

packages = [
    "mipac",
    "mipac.abstract",
    "mipac.actions",
    "mipac.actions.admins",
    "mipac.actions.drive",
    "mipac.actions.users",
    "mipac.errors",
    "mipac.manager",
    "mipac.manager.admins",
    "mipac.manager.drive",
    "mipac.manager.users",
    "mipac.models",
    "mipac.models.lite",
    "mipac.types",
    "mipac.utils",
]

setup(
    name="mipac",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    install_requires=requirements,
    url="https://github.com/yupix/mipac",
    author="yupix",
    author_email="yupi0982@outlook.jp",
    license="MIT",
    python_requires=">=3.12, <4.0",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=packages,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3.12",
        "Natural Language :: Japanese",
        "License :: OSI Approved :: MIT License",
    ],
    extras_require=extras_require,
)



from __future__ import annotations

from typing import TYPE_CHECKING, AsyncGenerator, Literal, overload, override

from mipac.abstract.action import AbstractAction
from mipac.http import HTTPClient, Route
from mipac.models.clip import Clip
from mipac.models.gallery import GalleryPost
from mipac.models.lite.user import PartialUser
from mipac.models.note import Note
from mipac.models.user import (
    Achievement,
    Follower,
    Following,
    FrequentlyRepliedUser,
    MeDetailed,
    UserDetailedNotMe,
    packed_user,
)
from mipac.types.clip import IClip
from mipac.types.follow import IFederationFollower, IFederationFollowing
from mipac.types.gallery import IGalleryPost
from mipac.types.note import INote
from mipac.types.user import (
    GetFrequentlyRepliedUsersResponse,
    IMeDetailedSchema,
    IUser,
    is_partial_user,
)
from mipac.utils.cache import cache
from mipac.utils.format import remove_dict_empty
from mipac.utils.pagination import Pagination

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager

__all__ = ["UserActions"]


class SharedUserActions(AbstractAction):
    def __init__(self, *, session: HTTPClient, client: ClientManager) -> None:
        self._session: HTTPClient = session
        self._client: ClientManager = client

    async def get_notes(
        self,
        with_replies: bool = False,
        with_renotes: bool = True,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        since_data: int | None = None,
        until_data: int | None = None,
        include_my_renotes: bool = True,
        with_files: bool = False,
        file_type: list[str] | None = None,
        exclude_nsfw: bool = False,
        *,
        user_id: str,
    ) -> list[Note]:  # TODO: since_dataなどを用いたページネーションを今後できるようにする
        data = {
            "userId": user_id,
            "withReplies": with_replies,
            "withRenotes": with_renotes,
            "limit": limit,
            "sinceId": since_id,
            "untilId": until_id,
            "sinceDate": since_data,
            "untilDate": until_data,
            "includeMyRenotes": include_my_renotes,
            "withFiles": with_files,
            "fileType": file_type,
            "excludeNsfw": exclude_nsfw,
        }

        raw_note: list[INote] = await self._session.request(
            Route("POST", "/api/users/notes"), json=data, auth=True
        )

        return [Note(raw_note=raw_note, client=self._client) for raw_note in raw_note]

    async def get_all_notes(
        self,
        with_replies: bool = False,
        with_renotes: bool = True,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        since_data: int | None = None,
        until_data: int | None = None,
        include_my_renotes: bool = True,
        with_files: bool = False,
        file_type: list[str] | None = None,
        exclude_nsfw: bool = False,
        *,
        user_id: str,
    ) -> AsyncGenerator[Note, None]:
        data = {
            "userId": user_id,
            "withReplies": with_replies,
            "withRenotes": with_renotes,
            "limit": limit,
            "sinceId": since_id,
            "untilId": until_id,
            "sinceDate": since_data,
            "untilDate": until_data,
            "includeMyRenotes": include_my_renotes,
            "withFiles": with_files,
            "fileType": file_type,
            "excludeNsfw": exclude_nsfw,
        }
        pagination = Pagination[INote](
            self._session, Route("POST", "/api/users/notes"), json=data, auth=True
        )

        while pagination.is_final is False:
            res_notes = await pagination.next()
            for note in res_notes:
                yield Note(note, client=self._client)

    async def get_clips(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        *,
        user_id: str,
    ) -> list[Clip]:
        data = {"userId": user_id, "limit": limit, "sinceId": since_id, "untilId": until_id}

        raw_clip: list[IClip] = await self._session.request(
            Route("POST", "/api/users/clips"), json=data, auth=True
        )

        return [Clip(raw_clip=raw_clip, client=self._client) for raw_clip in raw_clip]

    async def get_all_clips(
        self,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        *,
        user_id: str,
    ) -> AsyncGenerator[Clip, None]:
        data = {"userId": user_id, "limit": limit, "sinceId": since_id, "untilId": until_id}

        pagination = Pagination[IClip](
            self._session, Route("POST", "/api/users/clips"), json=data, auth=True
        )

        while pagination.is_final is False:
            clips: list[IClip] = await pagination.next()
            for clip in clips:
                yield Clip(raw_clip=clip, client=self._client)

    async def get_followers(
        self,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        username: str | None = None,
        host: str | None = None,
        *,
        user_id: str,
    ) -> list[Follower]:
        """
        Get followers of user.

        Endpoint: `/api/users/followers`

        Parameters
        ----------
        since_id : str, default=None
            Get followers after this id.
        until_id : str, default=None
            Get followers before this id.
        limit : int, default=10
            The maximum number of followers to return.
        username : str, default=None
            Get followers with this username.
        host : str, default=None
            Get followers with this host.
        user_id : str, default=None
            Get followers with this user id.

        Returns
        -------
        list[Follower]
            A list of followers.
        """
        data = {
            "userId": user_id,
            "sinceId": since_id,
            "untilId": until_id,
            "limit": limit,
            "username": username,
            "host": host,
        }
        raw_followers: list[IFederationFollower] = await self._session.request(
            Route("POST", "/api/users/followers"), json=data, auth=True
        )

        return [Follower(raw_follower, client=self._client) for raw_follower in raw_followers]

    async def get_all_followers(
        self,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        username: str | None = None,
        host: str | None = None,
        *,
        user_id: str,
    ) -> AsyncGenerator[Follower, None]:
        data = {
            "userId": user_id,
            "sinceId": since_id,
            "untilId": until_id,
            "limit": limit,
            "username": username,
            "host": host,
        }
        pagination = Pagination[IFederationFollower](
            self._session, Route("POST", "/api/users/followers"), json=data, auth=True
        )

        while pagination.is_final is False:
            raw_followers: list[IFederationFollower] = await pagination.next()
            for raw_follower in raw_followers:
                yield Follower(raw_follower, client=self._client)

    async def get_following(
        self,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        username: str | None = None,
        host: str | None = None,
        birthday: str | None = None,
        *,
        user_id: str,
    ) -> list[Following]:
        """Get following of user.

        Endpoint: `/api/users/following`

        Parameters
        ----------
        since_id : str, default=None
            Get following after this id.
        until_id : str, default=None
            Get following before this id.
        limit : int, default=10
            The maximum number of following to return.
        username : str, default=None
            Get following with this username.
        host : str, default=None
            Get following with this host.
        user_id : str, default=None
            Get following with this user id.

        Returns
        -------
        list[Following]
            A list of following.
        """
        data = {
            "userId": user_id,
            "sinceId": since_id,
            "untilId": until_id,
            "limit": limit,
            "username": username,
            "host": host,
            "birthday": birthday,
        }

        raw_following: list[IFederationFollowing] = await self._session.request(
            Route("POST", "/api/users/following"), json=data, auth=True
        )

        return [Following(raw_following, client=self._client) for raw_following in raw_following]

    async def get_all_following(
        self,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        username: str | None = None,
        host: str | None = None,
        birthday: str | None = None,
        *,
        user_id: str,
    ) -> AsyncGenerator[Following, None]:
        data = {
            "userId": user_id,
            "sinceId": since_id,
            "untilId": until_id,
            "limit": limit,
            "username": username,
            "host": host,
            "birthday": birthday,
        }

        pagination = Pagination[IFederationFollowing](
            self._session, Route("POST", "/api/users/following"), json=data, auth=True
        )

        while pagination.is_final is False:
            raw_followings: list[IFederationFollowing] = await pagination.next()
            for raw_following in raw_followings:
                yield Following(raw_following, client=self._client)

    async def get_gallery_posts(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        *,
        user_id: str,
    ) -> list[GalleryPost]:
        """Get gallery posts of user.

        Endpoint: `/api/users/gallery/posts`

        Parameters
        ----------
        limit : int, default=10
            The maximum number of gallery posts to return.
        since_id : str, default=None
            Get gallery posts after this id.
        until_id : str, default=None
            Get gallery posts before this id.
        user_id : str, default=None
            Get gallery posts with this user id.

        Returns
        -------
        list[GalleryPost]
            A list of gallery posts.
        """
        data = {
            "userId": user_id,
            "limit": limit,
            "sinceId": since_id,
            "untilId": until_id,
        }

        raw_gallery_posts: list[IGalleryPost] = await self._session.request(
            Route("POST", "/api/users/gallery/posts"), json=data, auth=True
        )

        return [
            GalleryPost(raw_gallery=raw_gallery, client=self._client)
            for raw_gallery in raw_gallery_posts
        ]

    async def get_all_gallery_posts(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        *,
        user_id: str,
    ) -> AsyncGenerator[GalleryPost, None]:
        data = {
            "userId": user_id,
            "limit": limit,
            "sinceId": since_id,
            "untilId": until_id,
        }

        pagination = Pagination[IGalleryPost](
            self._session, Route("POST", "/api/users/gallery/posts"), json=data, auth=True
        )

        while pagination.is_final is False:
            raw_gallery_posts: list[IGalleryPost] = await pagination.next()
            for raw_gallery_post in raw_gallery_posts:
                yield GalleryPost(raw_gallery=raw_gallery_post, client=self._client)

    async def get_frequently_replied_users(
        self, limit: int = 10, *, user_id: str
    ) -> list[FrequentlyRepliedUser]:
        """Get frequently replied users of user.

        Endpoint: `/api/users/get-frequently-replied-users`

        Parameters
        ----------
        limit : int, default=10
            The maximum number of frequently replied users to return.
        user_id : str, default=None
            Get frequently replied users with this user id.

        Returns
        -------
        list[FrequentlyRepliedUser]
            A list of frequently replied users.
        """
        data = {
            "userId": user_id,
            "limit": limit,
        }

        res: list[GetFrequentlyRepliedUsersResponse] = await self._session.request(
            Route("POST", "/api/users/get-frequently-replied-users"),
            json=data,
            auth=True,
            lower=True,
        )
        return [FrequentlyRepliedUser(i, client=self._client) for i in res]

    async def get_featured_notes(
        self, limit: int = 10, until_id: str | None = None, *, user_id: str
    ) -> list[Note]:
        """Get featured notes of user.

        Endpoint: `/api/users/featured-notes`

        Parameters
        ----------
        limit : int, default=10
            The maximum number of featured notes to return.
        until_id : str, default=None
            Get featured notes before this id.
        user_id : str, default=None
            Get featured notes with this user id.

        Returns
        -------
        list[Note]
            A list of featured notes.
        """
        data = {
            "userId": user_id,
            "limit": limit,
            "untilId": until_id,
        }

        raw_notes: list[INote] = await self._session.request(
            Route("POST", "/api/users/featured-notes"),
            json=data,
            auth=True,
            lower=True,
        )
        return [Note(raw_note=raw_note, client=self._client) for raw_note in raw_notes]

    async def get_all_featured_notes(
        self, limit: int = 10, until_id: str | None = None, *, user_id: str
    ) -> AsyncGenerator[Note, None]:
        """Get all featured notes of user.

        Endpoint: `/api/users/featured-notes`

        Parameters
        ----------
        limit : int, default=10
            The maximum number of featured notes to return.
        until_id : str, default=None
            Get featured notes before this id.
        user_id : str, default=None
            Get featured notes with this user id.

        Returns
        -------
        list[Note]
            A list of featured notes.
        """
        data = {
            "userId": user_id,
            "limit": limit,
            "untilId": until_id,
        }

        pagination = Pagination[INote](
            self._session, Route("POST", "/api/users/featured-notes"), json=data, auth=True
        )

        while pagination.is_final is False:
            raw_notes: list[INote] = await pagination.next()
            for raw_note in raw_notes:
                yield Note(raw_note=raw_note, client=self._client)

    async def get_achievements(self, *, user_id: str) -> list[Achievement]:
        """Get achievements of user."""
        data = {
            "userId": user_id,
        }
        res = await self._session.request(
            Route("POST", "/api/users/achievements"),
            json=data,
            auth=True,
            lower=True,
        )
        return [Achievement(i) for i in res]


class ClientUserActions(SharedUserActions):
    def __init__(self, user: PartialUser, *, session: HTTPClient, client: ClientManager):
        super().__init__(session=session, client=client)
        self.__user: PartialUser = user

    @override
    async def get_notes(
        self,
        with_replies: bool = False,
        with_renotes: bool = True,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        since_data: int | None = None,
        until_data: int | None = None,
        include_my_renotes: bool = True,
        with_files: bool = False,
        file_type: list[str] | None = None,
        exclude_nsfw: bool = False,
        *,
        user_id: str | None = None,
    ) -> list[Note]:  # TODO: since_dataなどを用いたページネーションを今後できるようにする
        user_id = user_id or self.__user.id

        return await super().get_notes(
            with_replies=with_replies,
            with_renotes=with_renotes,
            limit=limit,
            since_id=since_id,
            until_id=until_id,
            since_data=since_data,
            until_data=until_data,
            include_my_renotes=include_my_renotes,
            with_files=with_files,
            file_type=file_type,
            exclude_nsfw=exclude_nsfw,
            user_id=user_id,
        )

    @override
    async def get_all_notes(
        self,
        with_replies: bool = False,
        with_renotes: bool = True,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        since_data: int | None = None,
        until_data: int | None = None,
        include_my_renotes: bool = True,
        with_files: bool = False,
        file_type: list[str] | None = None,
        exclude_nsfw: bool = False,
        *,
        user_id: str | None = None,
    ) -> AsyncGenerator[Note, None]:
        user_id = user_id or self.__user.id

        async for i in super().get_all_notes(
            with_replies=with_replies,
            with_renotes=with_renotes,
            limit=limit,
            since_id=since_id,
            until_id=until_id,
            since_data=since_data,
            until_data=until_data,
            include_my_renotes=include_my_renotes,
            with_files=with_files,
            file_type=file_type,
            exclude_nsfw=exclude_nsfw,
            user_id=user_id,
        ):
            yield i

    @override
    async def get_clips(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        *,
        user_id: str | None = None,
    ) -> list[Clip]:
        user_id = user_id or self.__user.id

        return await super().get_clips(
            user_id=user_id, limit=limit, since_id=since_id, until_id=until_id
        )

    @override
    async def get_all_clips(
        self,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        *,
        user_id: str | None = None,
    ) -> AsyncGenerator[Clip, None]:
        user_id = user_id or self.__user.id

        async for i in super().get_all_clips(
            user_id=user_id, since_id=since_id, limit=limit, until_id=until_id
        ):
            yield i

    @override
    async def get_followers(
        self,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        username: str | None = None,
        host: str | None = None,
        *,
        user_id: str | None = None,
    ) -> list[Follower]:
        """
        Get followers of user.

        Endpoint: `/api/users/followers`

        Parameters
        ----------
        since_id : str, default=None
            Get followers after this id.
        until_id : str, default=None
            Get followers before this id.
        limit : int, default=10
            The maximum number of followers to return.
        username : str, default=None
            Get followers with this username.
        host : str, default=None
            Get followers with this host.
        user_id : str, default=None
            Get followers with this user id.

        Returns
        -------
        list[Follower]
            A list of followers.
        """
        user_id = user_id or self.__user.id

        return await super().get_followers(
            since_id=since_id,
            until_id=until_id,
            limit=limit,
            username=username,
            host=host,
            user_id=user_id,
        )

    @override
    async def get_all_followers(
        self,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        username: str | None = None,
        host: str | None = None,
        *,
        user_id: str | None = None,
    ) -> AsyncGenerator[Follower, None]:
        user_id = user_id or self.__user.id

        async for i in super().get_all_followers(
            since_id=since_id,
            until_id=until_id,
            limit=limit,
            username=username,
            host=host,
            user_id=user_id,
        ):
            yield i

    @override
    async def get_following(
        self,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        username: str | None = None,
        host: str | None = None,
        birthday: str | None = None,
        *,
        user_id: str | None = None,
    ) -> list[Following]:
        """Get following of user.

        Endpoint: `/api/users/following`

        Parameters
        ----------
        since_id : str, default=None
            Get following after this id.
        until_id : str, default=None
            Get following before this id.
        limit : int, default=10
            The maximum number of following to return.
        username : str, default=None
            Get following with this username.
        host : str, default=None
            Get following with this host.
        user_id : str, default=None
            Get following with this user id.

        Returns
        -------
        list[Following]
            A list of following.
        """
        user_id = user_id or self.__user.id

        return await super().get_following(
            since_id=since_id,
            until_id=until_id,
            limit=limit,
            username=username,
            host=host,
            birthday=birthday,
            user_id=user_id,
        )

    @override
    async def get_all_following(
        self,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        username: str | None = None,
        host: str | None = None,
        birthday: str | None = None,
        *,
        user_id: str | None = None,
    ) -> AsyncGenerator[Following, None]:
        user_id = user_id or self.__user.id

        async for i in super().get_all_following(
            since_id=since_id,
            until_id=until_id,
            limit=limit,
            username=username,
            host=host,
            birthday=birthday,
            user_id=user_id,
        ):
            yield i

    @override
    async def get_gallery_posts(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        *,
        user_id: str | None = None,
    ) -> list[GalleryPost]:
        """Get gallery posts of user.

        Endpoint: `/api/users/gallery/posts`

        Parameters
        ----------
        limit : int, default=10
            The maximum number of gallery posts to return.
        since_id : str, default=None
            Get gallery posts after this id.
        until_id : str, default=None
            Get gallery posts before this id.
        user_id : str, default=None
            Get gallery posts with this user id.

        Returns
        -------
        list[GalleryPost]
            A list of gallery posts.
        """
        user_id = user_id or self.__user.id

        return await super().get_gallery_posts(
            user_id=user_id, limit=limit, since_id=since_id, until_id=until_id
        )

    @override
    async def get_all_gallery_posts(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        *,
        user_id: str | None = None,
    ) -> AsyncGenerator[GalleryPost, None]:
        user_id = user_id or self.__user.id

        async for i in super().get_all_gallery_posts(
            user_id=user_id, limit=limit, since_id=since_id, until_id=until_id
        ):
            yield i

    @override
    async def get_frequently_replied_users(
        self, limit: int = 10, *, user_id: str | None = None
    ) -> list[FrequentlyRepliedUser]:
        """Get frequently replied users of user.

        Endpoint: `/api/users/get-frequently-replied-users`

        Parameters
        ----------
        limit : int, default=10
            The maximum number of frequently replied users to return.
        user_id : str, default=None
            Get frequently replied users with this user id.

        Returns
        -------
        list[FrequentlyRepliedUser]
            A list of frequently replied users.
        """
        user_id = user_id or self.__user.id

        return await super().get_frequently_replied_users(user_id=user_id)

    async def get_featured_notes(
        self, limit: int = 10, until_id: str | None = None, *, user_id: str | None = None
    ) -> list[Note]:
        """Get featured notes of user.

        Endpoint: `/api/users/featured-notes`

        Parameters
        ----------
        limit : int, default=10
            The maximum number of featured notes to return.
        until_id : str, default=None
            Get featured notes before this id.
        user_id : str, default=None
            Get featured notes with this user id.

        Returns
        -------
        list[Note]
            A list of featured notes.
        """
        user_id = user_id or self.__user.id

        return await super().get_featured_notes(user_id=user_id, limit=limit, until_id=until_id)

    async def get_all_featured_notes(
        self, limit: int = 10, until_id: str | None = None, *, user_id: str | None = None
    ) -> AsyncGenerator[Note, None]:
        """Get all featured notes of user.

        Endpoint: `/api/users/featured-notes`

        Parameters
        ----------
        limit : int, default=10
            The maximum number of featured notes to return.
        until_id : str, default=None
            Get featured notes before this id.
        user_id : str, default=None
            Get featured notes with this user id.

        Returns
        -------
        list[Note]
            A list of featured notes.
        """
        user_id = user_id or self.__user.id

        async for i in super().get_all_featured_notes(
            user_id=user_id, limit=limit, until_id=until_id
        ):
            yield i

    async def get_achievements(self, *, user_id: str | None = None) -> list[Achievement]:
        """Get achievements of user."""
        user_id = user_id or self.__user.id

        return await super().get_achievements(user_id=user_id)


class UserActions(SharedUserActions):
    def __init__(
        self,
        session: HTTPClient,
        client: ClientManager,
    ):
        super().__init__(session=session, client=client)

    async def get_me(self) -> MeDetailed:
        """
        ログインしているユーザーの情報を取得します
        """

        res: IMeDetailedSchema = await self._session.request(
            Route("POST", "/api/i"),
            lower=True,
        )
        return MeDetailed(res, client=self._client)

    @cache(group="get_user")
    async def get(
        self,
        user_id: str | None = None,
        user_ids: list[str] | None = None,
        username: str | None = None,
        host: str | None = None,
        **kwargs,
    ) -> UserDetailedNotMe | MeDetailed:
        """
        Retrieve user information from the user ID using the cache.
        If there is no cache, `fetch` is automatically used.
        The `fetch` method is recommended if you want up-to-date user information.

        Parameters
        ----------
        user_id : str
            target user id
        user_ids: list[str]
            target user ids
        username : str
            target username
        host : str, default=None
            Hosts with target users
        """

        field = remove_dict_empty(
            {"userId": user_id, "username": username, "host": host, "userIds": user_ids}
        )
        data: IUser = await self._session.request(
            Route("POST", "/api/users/show"), json=field, auth=True, lower=True
        )
        return packed_user(data, client=self._client)

    async def fetch(
        self,
        user_id: str | None = None,
        user_ids: list[str] | None = None,
        username: str | None = None,
        host: str | None = None,
    ) -> UserDetailedNotMe | MeDetailed:
        """
        Retrieve the latest user information using the target user ID or username.
        If you do not need the latest information, you should basically use the `get` method.
        This method accesses the server each time,
        which may increase the number of server accesses.

        Parameters
        ----------
        user_id : str
            target user id
        username : str
            target username
        username : str
            target username
        host : str, default=None
            Hosts with target users
        """
        return await self.get(
            user_id=user_id, username=username, host=host, user_ids=user_ids, cache_override=True
        )

    def get_mention(self, user: PartialUser) -> str:  # TODO: モデルに移す
        """
        Get mention name of user.

        Parameters
        ----------
        user : Optional[User], default=None
            The object of the user whose mentions you want to retrieve

        Returns
        -------
        str
            メンション
        """
        return f"@{user.username}@{user.host}" if user.instance else f"@{user.username}"

    async def search_by_username_and_host(
        self,
        username: str,
        host: str,
        limit: int = 100,
        detail: bool = True,
    ) -> list[UserDetailedNotMe | MeDetailed | PartialUser]:  # TODO: 続き
        """
        Search users by username and host.

        Parameters
        ----------
        username : str
            Username of user.
        host : str
            Host of user.
        limit : int, default=100
            The maximum number of users to return.
        detail : bool, default=True
            Weather to get detailed user information.

        Returns
        -------
        list[UserDetailedNotMe | MeDetailed | PartialUser]
            A list of users.
        """

        if limit > 100:
            raise ValueError("limit は100以下である必要があります")

        body = remove_dict_empty(
            {"username": username, "host": host, "limit": limit, "detail": detail}
        )
        res = await self._session.request(
            Route("POST", "/api/users/search-by-username-and-host"),
            lower=True,
            auth=True,
            json=body,
        )
        return [
            packed_user(user, client=self._client)
            if detail
            else PartialUser(user, client=self._client)
            for user in res
        ]

    @overload
    async def search(
        self,
        query: str,
        limit: int = 100,
        offset: int = 0,
        origin: Literal["local", "remote", "combined"] = "combined",
        detail: Literal[False] = ...,
        *,
        get_all: bool = False,
    ) -> AsyncGenerator[PartialUser, None]:
        ...

    @overload
    async def search(
        self,
        query: str,
        limit: int = 100,
        offset: int = 0,
        origin: Literal["local", "remote", "combined"] = "combined",
        detail: Literal[True] = True,
        *,
        get_all: bool = False,
    ) -> AsyncGenerator[UserDetailedNotMe | MeDetailed, None]:
        ...

    async def search(
        self,
        query: str,
        limit: int = 100,
        offset: int = 0,
        origin: Literal["local", "remote", "combined"] = "combined",
        detail: Literal[True, False] = True,
        *,
        get_all: bool = False,
    ) -> AsyncGenerator[UserDetailedNotMe | MeDetailed | PartialUser, None]:
        """
        Search users by keyword.

        Parameters
        ----------
        query : str
            Keyword to search.
        limit : int, default=100
            The maximum number of users to return.
        offset : int, default=0
            The number of users to skip.
        origin : Literal['local', 'remote', 'combined'], default='combined'
            The origin of users to search.
        detail : Literal[True, False], default=True
            Whether to return detailed user information.
        get_all : bool, default=False
            Whether to return all users.

        """

        if limit > 100:
            raise ValueError("limit は100以下である必要があります")

        if get_all:
            limit = 100

        body = remove_dict_empty(
            {"query": query, "limit": limit, "offset": offset, "origin": origin, "detail": detail}
        )

        pagination = Pagination[IUser](
            self._session,
            Route("POST", "/api/users/search"),
            json=body,
            pagination_type="count",
        )

        while True:
            users: list[IUser] = await pagination.next()
            for user in users:
                yield (
                    packed_user(user, client=self._client)
                    if is_partial_user(user) is False
                    else PartialUser(user, client=self._client)
                )
            if get_all is False or pagination.is_final:
                break



from __future__ import annotations

from typing import TYPE_CHECKING, Any, AsyncGenerator

from mipac.abstract.action import AbstractAction
from mipac.http import HTTPClient, Route
from mipac.models.instance import FederationInstance
from mipac.models.user import MeDetailed, UserDetailedNotMe, packed_user
from mipac.types.follow import IFederationFollower, IFederationFollowing
from mipac.types.instance import IFederationInstance, IFederationInstanceStat
from mipac.types.user import IUserDetailed
from mipac.utils.pagination import Pagination

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class FederationActions(AbstractAction):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        self.__session = session
        self.__client = client

    async def get_ap(self, uri: str) -> dict[Any, Any]:
        return dict(
            await self.__session.request(
                Route("POST", "/api/ap/get"), auth=True, json={"uri": uri}, lower=True
            )
        )

    async def show_ap(
        self, host: str, since_id: str | None = None, until_id: str | None = None, limit: int = 10
    ) -> FederationInstance:  # TODO: 存在するのか確認する
        body = {"host": host, "sinceId": since_id, "untilId": until_id, "limit": limit}

        res: FederationInstance = await self.__session.request(
            Route("POST", "/api/ap/show"), auth=True, json=body
        )
        return res

    async def get_followers(
        self,
        host: str,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        get_all: bool = False,
    ):
        if limit > 100:
            raise ValueError("limitは100以下である必要があります")

        if get_all:
            limit = 100

        body = {"host": host, "sinceId": since_id, "untilId": until_id, "limit": limit}

        pagination = Pagination[IFederationFollower](
            self.__session, Route("POST", "/api/federation/followers"), json=body
        )

        while True:
            res_federation_followers: list[IFederationFollower] = await pagination.next()
            for federation_follower in res_federation_followers:
                yield federation_follower

            if get_all is False or pagination.is_final:
                break

    async def get_following(
        self,
        host: str,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        get_all: bool = False,
    ):
        if limit > 100:
            raise ValueError("limitは100以下である必要があります")

        if get_all:
            limit = 100

        body = {"host": host, "sinceId": since_id, "untilId": until_id, "limit": limit}

        pagination = Pagination[IFederationFollowing](
            self.__session, Route("POST", "/api/federation/following"), json=body
        )

        while True:
            res_federation_followings: list[IFederationFollowing] = await pagination.next()
            for federation_following in res_federation_followings:
                yield federation_following

            if get_all is False or pagination.is_final:
                break

    async def get_instances(
        self,
        host: str | None = None,
        blocked: bool | None = None,
        not_responding: bool | None = None,
        suspended: bool | None = None,
        federating: bool | None = None,
        subscribing: bool | None = None,
        publishing: bool | None = None,
        limit: int = 30,
        offset: int = 0,
        sort: str | None = None,
    ) -> list[FederationInstance]:
        if limit > 100:
            raise ValueError("limitは100以下である必要があります")
        body = {
            "host": host,
            "blocked": blocked,
            "notResponding": not_responding,
            "suspended": suspended,
            "federating": federating,
            "subscribing": subscribing,
            "publishing": publishing,
            "limit": limit,
            "offset": offset,
            "sort": sort,
        }

        res: list[IFederationInstance] = await self.__session.request(
            Route("POST", "/api/federation/instances"), auth=True, lower=True, json=body
        )

        return [FederationInstance(i, client=self.__client) for i in res]

    async def show_instance(self, host: str) -> FederationInstance:
        res: IFederationInstance = await self.__session.request(
            Route("POST", "/api/federation/show-instance"),
            auth=True,
            json={"host": host},
            lower=True,
        )

        return FederationInstance(res, client=self.__client)

    async def update_remote_user(self, user_id: str) -> bool:
        return bool(
            self.__session.request(
                Route("POST", "/api/federation/update-remote-user"),
                auth=True,
                json={"userId": user_id},
            )
        )

    async def get_users(
        self,
        host: str,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        get_all: bool = False,
    ) -> AsyncGenerator[UserDetailedNotMe | MeDetailed, None]:
        if limit > 100:
            raise ValueError("limitは100以下である必要があります")

        if get_all:
            limit = 100

        body = {"host": host, "sinceId": since_id, "untilId": until_id, "limit": limit}

        pagination = Pagination[IUserDetailed](
            self.__session, Route("POST", "/api/federation/users"), json=body
        )

        while True:
            res_users: list[IUserDetailed] = await pagination.next()
            for user in res_users:
                yield packed_user(user, client=self.__client)

            if get_all is False or pagination.is_final:
                break

    async def get_stats(self, limit: int = 10) -> IFederationInstanceStat:
        if limit > 100:
            raise ValueError("limitは100以下である必要があります")
        res: IFederationInstanceStat = await self.__session.request(
            Route("POST", "/api/federation/stats"), auth=True, body={"limit": limit}, lower=True
        )
        return res



from typing import NotRequired, TypedDict

from mipac.types.note import INote


class IChannel(TypedDict):
    id: str
    created_at: str
    last_noted_at: str | None
    name: str
    description: str | None
    user_id: str | None
    banner_url: str | None
    pinned_note_ids: list[str]
    color: str
    is_archived: bool
    users_count: int
    notes_count: int
    is_sensitive: bool
    allow_renote_to_external: bool
    is_following: NotRequired[bool]
    is_favorited: NotRequired[bool]
    pinned_notes: NotRequired[list[INote]]



from typing import NotRequired, TypedDict

from mipac.types.user import IPartialUser, IUserDetailedNotMeSchema


class IFederationFollowCommon(TypedDict):
    id: str
    created_at: str
    followee_id: str
    followee: NotRequired[IUserDetailedNotMeSchema]
    follower_id: str
    follower: NotRequired[IUserDetailedNotMeSchema]


class IFederationFollower(IFederationFollowCommon):
    ...


class IFederationFollowing(IFederationFollowCommon):
    ...


class IFollowRequest(TypedDict):
    id: str
    follower: IPartialUser
    followee: IPartialUser



from __future__ import annotations

from typing import TYPE_CHECKING, AsyncGenerator, override

from mipac.abstract.action import AbstractAction
from mipac.http import HTTPClient, Route
from mipac.models.antenna import Antenna
from mipac.models.note import Note
from mipac.types.antenna import IAntenna, IAntennaReceiveSource
from mipac.types.note import INote
from mipac.utils.format import remove_dict_empty, remove_dict_missing
from mipac.utils.pagination import Pagination
from mipac.utils.util import MISSING

if TYPE_CHECKING:
    from mipac.client import ClientManager


class SharedAntennaActions(AbstractAction):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        self._session: HTTPClient = session
        self._client: ClientManager = client

    async def delete(self, *, antenna_id: str) -> bool:
        """
        Delete antenna from identifier

        Parameters
        ----------
        antenna_id : str | None, optional
            target identifier

        Returns
        -------
        bool
            success or failure
        """

        body = {"antennaId": antenna_id}
        res: bool = await self._session.request(
            Route("POST", "/api/antennas/delete"), auth=True, json=body
        )
        return res

    async def show(self, *, antenna_id: str) -> Antenna:
        """Show antenna from identifier

        Parameters
        ----------
        antenna_id : str | None, optional
            target identifier, by default None

        Returns
        -------
        Antenna
            antenna object
        """
        body = {"antennaId": antenna_id}
        res_antenna: IAntenna = await self._session.request(
            Route("POST", "/api/antennas/show"), auth=True, json=body
        )
        return Antenna(res_antenna, client=self._client)

    async def get_notes(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        since_date: str | None = None,
        until_date: str | None = None,
        *,
        antenna_id: str,
    ) -> list[Note]:
        body = remove_dict_empty(
            {
                "antennaId": antenna_id,
                "limit": limit,
                "sinceId": since_id,
                "untilId": until_id,
                "sinceDate": since_date,
                "untilDate": until_date,
            }
        )

        res: list[INote] = await self._session.request(
            Route("POST", "/api/antennas/notes"), json=body
        )
        return [Note(note, client=self._client) for note in res]

    async def get_all_notes(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        since_date: str | None = None,
        until_date: str | None = None,
        *,
        antenna_id: str,
    ) -> AsyncGenerator[Note, None]:
        body = remove_dict_empty(
            {
                "antennaId": antenna_id,
                "limit": limit,
                "sinceId": since_id,
                "untilId": until_id,
                "sinceDate": since_date,
                "untilDate": until_date,
            }
        )

        pagination = Pagination[INote](
            self._session, Route("POST", "/api/antennas/notes"), json=body
        )

        while pagination.is_final is False:
            res_notes = await pagination.next()
            for res_note in res_notes:
                yield Note(res_note, client=self._client)

    async def update(
        self,
        name: str,
        src: IAntennaReceiveSource,
        keywords: list[list[str]],
        exclude_keywords: list[list[str]],
        users: list[str],
        case_sensitive: bool,
        with_replies: bool,
        with_file: bool,
        notify: bool,
        user_list_id: str | None = None,
        *,
        antenna_id: str,
    ) -> Antenna:
        """Update an antenna.

        Parameters
        ----------
        name : str
            Name of the antenna.
        src : IAntennaReceiveSource
            Receive source of the antenna.
        keywords : list[list[str]]
            Receive keywords.
        exclude_keywords : list[list[str]]
            Excluded keywords.
        users : list[str]
            List of target user ID. Required when selecting 'users' as the receive source.
        case_sensitive : bool
            Whether to differentiate between uppercase and lowercase letters.
        with_replies : bool
            Whether to include replies.
        with_file : bool
            Whether to limit to notes with attached files.
        notify : bool
            Whether to notify for new notes.
        user_list_id : str | None, default None
            List of user IDs when selecting 'users' as the receive source for the antenna.

        Returns
        -------
        Antenna
            The created antenna.
        """
        if (
            all(
                [
                    name,
                    src,
                    keywords,
                ]
            )
            is False
        ):
            raise ValueError("Required parameters are missing")
        body = {
            "antennaId": antenna_id,
            "name": name,
            "src": src,
            "userListId": user_list_id,
            "keywords": keywords,
            "excludeKeywords": exclude_keywords,
            "users": users,
            "caseSensitive": case_sensitive,
            "withReplies": with_replies,
            "withFile": with_file,
            "notify": notify,
        }

        res_antenna: IAntenna = await self._session.request(
            Route("POST", "/api/antennas/update"), auth=True, json=body, remove_none=False
        )

        return Antenna(res_antenna, client=self._client)


class ClientAntennaActions(SharedAntennaActions):
    def __init__(self, *, antenna_id: str, session: HTTPClient, client: ClientManager):
        super().__init__(session=session, client=client)
        self.__antenna_id: str = antenna_id

    @override
    async def delete(self, *, antenna_id: str | None = None) -> bool:
        """
        Delete antenna from identifier

        Parameters
        ----------
        antenna_id : str | None, optional
            target identifier

        Returns
        -------
        bool
            success or failure
        """
        antenna_id = antenna_id or self.__antenna_id

        return await super().delete(antenna_id=antenna_id)

    @override
    async def show(self, *, antenna_id: str | None = None) -> Antenna:
        """Show antenna from identifier

        Parameters
        ----------
        antenna_id : str | None, optional
            target identifier, by default None

        Returns
        -------
        Antenna
            antenna object
        """
        antenna_id = antenna_id or self.__antenna_id

        return await super().show(antenna_id=antenna_id)

    @override
    async def get_notes(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        since_date: str | None = None,
        until_date: str | None = None,
        *,
        antenna_id: str | None = None,
    ) -> list[Note]:
        antenna_id = antenna_id or self.__antenna_id

        return await super().get_notes(
            limit=limit,
            since_id=since_id,
            until_id=until_id,
            since_date=since_date,
            until_date=until_date,
            antenna_id=antenna_id,
        )

    @override
    async def get_all_notes(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        since_date: str | None = None,
        until_date: str | None = None,
        *,
        antenna_id: str | None = None,
    ) -> AsyncGenerator[Note, None]:
        antenna_id = antenna_id or self.__antenna_id

        async for i in super().get_all_notes(
            limit, since_id, until_id, since_date, until_date, antenna_id=antenna_id
        ):
            yield i

    @override
    async def update(
        self,
        name: str,
        src: IAntennaReceiveSource,
        keywords: list[list[str]],
        exclude_keywords: list[list[str]],
        users: list[str],
        case_sensitive: bool,
        with_replies: bool,
        with_file: bool,
        notify: bool,
        user_list_id: str | None = None,
        *,
        antenna_id: str | None = None,
    ) -> Antenna:
        """Update an antenna.

        Parameters
        ----------
        name : str
            Name of the antenna.
        src : IAntennaReceiveSource
            Receive source of the antenna.
        keywords : list[list[str]]
            Receive keywords.
        exclude_keywords : list[list[str]]
            Excluded keywords.
        users : list[str]
            List of target user ID. Required when selecting 'users' as the receive source.
        case_sensitive : bool
            Whether to differentiate between uppercase and lowercase letters.
        with_replies : bool
            Whether to include replies.
        with_file : bool
            Whether to limit to notes with attached files.
        notify : bool
            Whether to notify for new notes.
        user_list_id : str | None, default None
            List of user IDs when selecting 'users' as the receive source for the antenna.

        Returns
        -------
        Antenna
            The created antenna.
        """

        antenna_id = antenna_id or self.__antenna_id

        return await super().update(
            name=name,
            src=src,
            keywords=keywords,
            exclude_keywords=exclude_keywords,
            users=users,
            case_sensitive=case_sensitive,
            with_replies=with_replies,
            with_file=with_file,
            notify=notify,
            user_list_id=user_list_id,
            antenna_id=antenna_id,
        )


class AntennaActions(SharedAntennaActions):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        super().__init__(session=session, client=client)

    async def create(
        self,
        name: str,
        src: IAntennaReceiveSource,
        keywords: list[list[str]],
        exclude_keywords: list[list[str]] | None = None,
        users: list[str] | None = None,
        case_sensitive: bool = False,
        local_only: bool = MISSING,
        with_replies: bool = False,
        with_file: bool = False,
        notify: bool = False,
        user_list_id: str | None = None,
    ) -> Antenna:
        """Create an antenna.

        Parameters
        ----------
        name : str
            Name of the antenna.
        src : IAntennaReceiveSource
            Receive source of the antenna.
        keywords : list[list[str]]
            Receive keywords.
        exclude_keywords : list[list[str]] | None, default None
            Excluded keywords.
        local_only : bool, default MISSING
            Whether to limit to local notes.
        users : list[str] | None, default None
            List of target user ID. Required when selecting 'users' as the receive source.
        case_sensitive : bool, default False
            Whether to differentiate between uppercase and lowercase letters.
        with_replies : bool, default False
            Whether to include replies.
        with_file : bool, default False
            Whether to limit to notes with attached files.
        notify : bool, default False
            Whether to notify for new notes.
        user_list_id : str | None, default None
            List of user IDs when selecting 'users' as the receive source for the antenna.

        Returns
        -------
        Antenna
            The created antenna.
        """

        if users is None:
            users = [""]
        if exclude_keywords is None:
            exclude_keywords = [[]]

        if (
            all(
                [
                    name,
                    src,
                    keywords,
                ]
            )
            is False
        ):
            raise ValueError("Required parameters are missing")
        body = remove_dict_missing(
            {
                "name": name,
                "src": src,
                "userListId": user_list_id,
                "keywords": keywords,
                "excludeKeywords": exclude_keywords,
                "users": users,
                "caseSensitive": case_sensitive,
                "localOnly": local_only,
                "withReplies": with_replies,
                "withFile": with_file,
                "notify": notify,
            }
        )

        res_antenna: IAntenna = await self._session.request(
            Route("POST", "/api/antennas/create"), auth=True, json=body, remove_none=False
        )

        return Antenna(res_antenna, client=self._client)

    async def get_list(self) -> list[Antenna]:
        res_antennas: list[IAntenna] = await self._session.request(
            Route("POST", "/api/antennas/list"), auth=True
        )

        return [Antenna(res_antenna, client=self._client) for res_antenna in res_antennas]



from __future__ import annotations

import io
import os
from typing import TYPE_CHECKING, Any, AsyncGenerator, override

from mipac.abstract.action import AbstractAction
from mipac.http import HTTPClient, Route
from mipac.models.drive import File
from mipac.models.note import Note
from mipac.types.drive import IDriveSort, IFile
from mipac.types.note import INote
from mipac.utils.format import bool_to_string, remove_dict_missing
from mipac.utils.pagination import Pagination
from mipac.utils.util import MISSING

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class SharedFileActions(AbstractAction):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        self._session: HTTPClient = session
        self._client: ClientManager = client

    async def get_attached_notes(
        self,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        *,
        file_id: str,
    ) -> list[Note]:
        """Get the attached notes of a file

        Endpoint: `/api/drive/files/attached-notes`

        Parameters
        ----------
        since_id: str | None
            The id of the note to start from, defaults to None
        until_id: str | None
            The id of the note to end at, defaults to None
        limit: int
            The amount of notes to get, defaults to 10
        file_id: str | None
            The id of the file to get notes from, defaults to None

        Returns
        -------
        list[Note]
            The attached notes of the file
        """
        body = {
            "sinceId": since_id,
            "untilId": until_id,
            "limit": limit,
            "fileId": file_id,
        }

        raw_notes: list[INote] = await self._session.request(
            Route("POST", "/api/drive/files/attached-notes"), json=body, auth=True
        )
        return [Note(raw_note, client=self._client) for raw_note in raw_notes]

    async def get_all_attached_notes(
        self,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        *,
        file_id: str,
    ) -> AsyncGenerator[Note, None]:
        body = {
            "sinceId": since_id,
            "untilId": until_id,
            "limit": limit,
            "fileId": file_id,
        }

        pagination = Pagination[INote](
            self._session,
            Route("POST", "/api/drive/files/attached-notes"),
            json=body,
            auth=True,
        )

        while pagination.is_final is False:
            for raw_note in await pagination.next():
                yield Note(raw_note, client=self._client)

    async def delete(self, *, file_id: str) -> bool:
        """指定したファイルIDのファイルを削除します

        Endpoint: `/api/drive/files/delete`

        Parameters
        ----------
        file_id: str | None
            対象のファイルID, default=None

        Returns
        -------
        bool
            削除に成功したかどうか
        """
        data = {"fileId": file_id}

        res: bool = await self._session.request(
            Route("POST", "/api/drive/files/delete"), json=data, auth=True
        )
        return res

    async def update(
        self,
        folder_id: str | None = MISSING,
        name: str | None = MISSING,
        is_sensitive: bool = MISSING,
        comment: str | None = MISSING,
        *,
        file_id: str,
    ) -> File:
        """指定したIDのファイル情報を更新します

        Endpoint: `/api/drive/files/update`

        Parameters
        ----------
        folder_id: str | None
            ファイルを置くフォルダID, default=MISSING
        name: str | None
            ファイル名, default=MISSING
        is_sensitive: bool
            ファイルがセンシティブかどうか, default=MISSING
        comment: str | None
            ファイルのコメント, default=MISSING
        file_id: str | None
            対象のファイルID, default=None

        Returns
        -------
        File
            更新後のファイル
        """
        data = remove_dict_missing(
            {
                "fileId": file_id,
                "folderId": folder_id,
                "name": name,
                "isSensitive": is_sensitive,
                "comment": comment,
            }
        )

        res: IFile = await self._session.request(
            Route("POST", "/api/drive/files/update"), json=data, auth=True
        )
        return File(res, client=self._client)


class ClientFileActions(SharedFileActions):
    def __init__(self, file_ids: str, *, session: HTTPClient, client: ClientManager):
        super().__init__(session=session, client=client)
        self.__file_ids: str = file_ids

    @override
    async def get_attached_notes(
        self,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        *,
        file_id: str | None = None,
    ) -> list[Note]:
        """Get the attached notes of a file

        Endpoint: `/api/drive/files/attached-notes`

        Parameters
        ----------
        since_id: str | None
            The id of the note to start from, defaults to None
        until_id: str | None
            The id of the note to end at, defaults to None
        limit: int
            The amount of notes to get, defaults to 10
        file_id: str | None
            The id of the file to get notes from, defaults to None

        Returns
        -------
        list[Note]
            The attached notes of the file
        """

        file_id = file_id or self.__file_ids

        return await super().get_attached_notes(
            since_id=since_id, until_id=until_id, limit=limit, file_id=file_id
        )

    @override
    async def get_all_attached_notes(
        self,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 10,
        *,
        file_id: str | None = None,
    ) -> AsyncGenerator[Note, None]:
        file_id = file_id or self.__file_ids

        async for note in super().get_all_attached_notes(
            since_id=since_id, until_id=until_id, limit=limit, file_id=file_id
        ):
            yield note

    @override
    async def delete(self, *, file_id: str | None = None) -> bool:
        """指定したファイルIDのファイルを削除します

        Endpoint: `/api/drive/files/delete`

        Parameters
        ----------
        file_id: str | None
            対象のファイルID, default=None

        Returns
        -------
        bool
            削除に成功したかどうか
        """

        file_id = file_id or self.__file_ids

        return await super().delete(file_id=file_id)

    @override
    async def update(
        self,
        folder_id: str | None = MISSING,
        name: str | None = MISSING,
        is_sensitive: bool = MISSING,
        comment: str | None = MISSING,
        *,
        file_id: str | None = None,
    ) -> File:
        """指定したIDのファイル情報を更新します

        Endpoint: `/api/drive/files/update`

        Parameters
        ----------
        folder_id: str | None
            ファイルを置くフォルダID, default=MISSING
        name: str | None
            ファイル名, default=MISSING
        is_sensitive: bool
            ファイルがセンシティブかどうか, default=MISSING
        comment: str | None
            ファイルのコメント, default=MISSING
        file_id: str | None
            対象のファイルID, default=None

        Returns
        -------
        File
            更新後のファイル
        """
        file_id = file_id or self.__file_ids

        return await super().update(
            folder_id=folder_id,
            name=name,
            is_sensitive=is_sensitive,
            comment=comment,
            file_id=file_id,
        )


class FileActions(SharedFileActions):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        super().__init__(session=session, client=client)

    async def get_files(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        folder_id: str | None = None,
        type: str | None = None,
        sort: IDriveSort | None = None,
    ) -> list[File]:
        """Get the files of the drive

        Endpoint: `/api/drive/files`

        Parameters
        ----------
        limit: int
            The amount of files to get, defaults to 10
        since_id: str | None
            The id of the file to start from, defaults to None
        until_id: str | None
            The id of the file to end at, defaults to None
        folder_id: str | None
            The id of the folder to get files from, defaults to None
        type: str | None
            The type of file to get, defaults to None
        sort: IDriveSort | None
            The way to sort the files, defaults to None

        Returns
        -------
        list[File]
            The files of the drive
        """

        data = {
            "limit": limit,
            "sinceId": since_id,
            "untilId": until_id,
            "folderId": folder_id,
            "type": type,
            "sort": sort,
        }

        raw_files: list[IFile] = await self._session.request(
            Route("POST", "/api/drive/files"), json=data, auth=True
        )
        return [File(raw_file, client=self._client) for raw_file in raw_files]

    async def get_all_files(
        self,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        folder_id: str | None = None,
        type: str | None = None,
        sort: IDriveSort | None = None,
    ) -> AsyncGenerator[File, None]:
        body = {
            "limit": limit,
            "sinceId": since_id,
            "untilId": until_id,
            "folderId": folder_id,
            "type": type,
            "sort": sort,
        }

        pagination = Pagination[IFile](
            self._session, Route("POST", "/api/drive/files"), json=body, auth=True
        )

        while pagination.is_final is False:
            for raw_file in await pagination.next():
                yield File(raw_file, client=self._client)

    async def check_existence(self, md5: str) -> bool:
        """指定したmd5のファイルが既に存在するか確認します

        Endpoint: `/api/drive/files/check-existence`

        Parameters
        ----------
        md5: str
            確認したいmd5

        Returns
        -------
        bool
            存在するかしないか
        """

        data = {"md5": md5}

        res: bool = await self._session.request(
            Route("POST", "/api/drive/files/check-existence"), json=data, auth=True
        )
        return res

    async def create(
        self,
        file: str | bytes | os.PathLike[Any] | io.BufferedIOBase,
        folder_id: str | None = None,
        name: str | None = None,
        comment: str | None = None,
        is_sensitive: bool = False,
        force: bool = False,
    ) -> File:
        """Upload a file to the drive

        Endpoint: `/api/drive/files/create`

        Parameters
        ----------
        file: str | bytes | os.PathLike[Any] | io.BufferedIOBase
            アップロードするファイル
        folder_id: str | None
            アップロード先のフォルダID, default=None
        name: str | None
            ファイルの名前, default=None
        comment: str | None
            ファイルのコメント, default=None
        is_sensitive: bool
            ファイルがセンシティブかどうか, default=False
        force: bool
            ファイルが既に存在する場合でも強制的にアップロードするかどうか, default=False


        Returns
        -------
        File
            アップロードしたファイル
        """

        if isinstance(file, io.IOBase):
            if (file.seekable() and file.readable()) is False:  # 書き込み/読み込みができるか確認
                raise ValueError(f"File buffer {file!r} must be seekable and readable")
            file_byte = file
        elif isinstance(file, bytes):
            file_byte = io.BytesIO(file)
        else:
            file_byte = open(file, "rb")

        data = {
            "folderId": folder_id,
            "name": name,
            "comment": comment,
            "isSensitive": bool_to_string(is_sensitive),
            "force": bool_to_string(force),
            "file": file_byte,
        }
        res: IFile = await self._session.request(
            Route("POST", "/api/drive/files/create"),
            data=data,
            auth=True,
            lower=True,
        )
        return File(res, client=self._client)

    async def find_by_hash(self, md5: str) -> list[File]:
        """指定したハッシュのファイルを検索します

        Endpoint: `/api/drive/files/find-by-hash`

        Parameters
        ----------
        md5: str
            検索したいファイルのハッシュ

        Returns
        -------
        list[File]
            見つかったファイル
        """

        data = {"md5": md5}

        raw_files: list[IFile] = await self._session.request(
            Route("POST", "/api/drive/files/find-by-hash"), json=data, auth=True
        )
        return [File(raw_file, client=self._client) for raw_file in raw_files]

    async def find(self, name: str, folder_id: str | None = None) -> list[File]:
        """指定した名前のファイルを検索します

        Endpoint: `/api/drive/files/find`

        Parameters
        ----------
        name: str
            検索したいファイルの名前
        folder_id: str | None
            ファイルを検索するフォルダID, default=None

        Returns
        -------
        list[File]
            The found files
        """

        data = {"name": name, "folderId": folder_id}

        res: list[IFile] = await self._session.request(
            Route("POST", "/api/drive/files/find"), json=data, auth=True
        )
        return [File(raw_file, client=self._client) for raw_file in res]

    async def show(self, file_id: str, url: str | None = None) -> File:
        """指定したIDのファイル情報を取得します

        Endpoint: `/api/drive/files/show`

        Parameters
        ----------
        file_id: str
            対象のファイルID
        url: str | None
            取得したいファイルのURL, default=None

        Returns
        -------
        File
            取得したファイル
        """

        data = {"fileId": file_id, "url": url}

        res: IFile = await self._session.request(
            Route("POST", "/api/drive/files/show"), json=data, auth=True
        )
        return File(res, client=self._client)

    async def upload_from_url(
        self,
        url: str,
        folder_id: str | None = None,
        is_sensitive: bool = False,
        comment: str | None = None,
        marker: str | None = None,
        force: bool = False,
    ):
        """Upload a file to the drive from a url

        Endpoint: `/api/drive/files/upload-from-url`

        Parameters
        ----------
        url: str
            アップロードするファイルのURL
        folder_id: str | None
            アップロード先のフォルダID, default=None
        is_sensitive: bool
            ファイルがセンシティブかどうか, default=False
        comment: str | None
            ファイルのコメント, default=None
        marker: str | None
            ストリーミング通信でアップロード完了後に区別するためのマーカー, default=None
        force: bool
            同様のファイルが既に存在する場合でも強制的にアップロードするかどうか, default=False

        Returns
        -------
        bool
            アップロードのリクエストに成功したかどうか
        """

        data = {
            "url": url,
            "folderId": folder_id,
            "isSensitive": is_sensitive,
            "comment": comment,
            "marker": marker,
            "force": force,
        }

        res: bool = await self._session.request(
            Route("POST", "/api/drive/files/upload-from-url"), json=data, auth=True
        )
        return res



from __future__ import annotations

from typing import TYPE_CHECKING, Any, AsyncGenerator, Literal

from mipac.abstract.action import AbstractAction
from mipac.http import Route
from mipac.models.roles import Role, RoleUser
from mipac.types.meta import IPolicies
from mipac.types.roles import IRole, IRoleUser
from mipac.utils.pagination import Pagination

if TYPE_CHECKING:
    from mipac.http import HTTPClient
    from mipac.manager.client import ClientManager


class AdminRoleModelActions(AbstractAction):
    def __init__(self, role_id: str | None = None, *, session: HTTPClient, client: ClientManager):
        self._session: HTTPClient = session
        self._client: ClientManager = client
        self._role_id: str | None = role_id

    async def update(
        self,
        name: str,
        description: str,
        role_id: str | None = None,
        color: str | None = None,
        iconUrl: str | None = None,
        target: Literal["manual", "conditional"] = "manual",
        cond_formula: dict[Any, Any] | None = None,
        is_public: bool = False,
        is_moderator: bool = False,
        is_administrator: bool = False,
        as_badge: bool = False,
        can_edit_members_by_moderator: bool = False,
        policies: dict[Any, Any] | None = None,
    ) -> bool:
        role_id = self._role_id or role_id
        if role_id is None:
            raise ValueError("role_idは必須です")
        body = {
            "roleId": role_id,
            "name": name,
            "description": description,
            "color": color,
            "iconUrl": iconUrl,
            "target": target,
            "condFormula": cond_formula or {},
            "isPublic": is_public,
            "isModerator": is_moderator,
            "isAdministrator": is_administrator,
            "asBadge": as_badge,
            "canEditMembersByModerator": can_edit_members_by_moderator,
            "policies": policies or {},
        }
        res: bool = await self._session.request(
            Route("POST", "/api/admin/roles/update"),
            json=body,
            auth=True,
            lower=True,
            remove_none=False,
        )
        return res

    async def delete(self, role_id: str | None = None) -> bool:
        role_id = self._role_id or role_id
        if role_id is None:
            raise ValueError("role_idは必須です")
        res: bool = await self._session.request(
            Route("POST", "/api/admin/roles/delete"),
            auth=True,
            json={"roleId": role_id},
            lower=True,
        )
        return res

    async def assign(
        self, user_id: str, role_id: str | None = None, expires_at: int | None = None
    ) -> bool:
        """指定したユーザーに指定したロールを付与します

        Parameters
        ----------
        role_id : str
            ロールのID
        user_id : str
            ロールを付与する対象のユーザーID
        expires_at : int | None, optional
            いつまでロールを付与するか, by default None

        Returns
        -------
        bool
            成功したか否か
        """
        if role_id is None:
            raise ValueError("role_idは必須です")
        body = {"roleId": role_id, "userId": user_id, "expiresAt": expires_at}
        res: bool = await self._session.request(
            Route("POST", "/api/admin/roles/assign"), auth=True, json=body
        )
        return res

    async def unassign(self, user_id: str, role_id: str | None = None) -> bool:
        """指定したユーザーに指定したロールを付与します

        Parameters
        ----------
        role_id : str
            ロールのID
        user_id : str
            ロールを付与する対象のユーザーID
        expires_at : int | None, optional
            いつまでロールを付与するか, by default None

        Returns
        -------
        bool
            成功したか否か
        """
        role_id = self._role_id or role_id
        if role_id is None:
            raise ValueError("role_idは必須です")
        body = {"roleId": role_id, "userId": user_id}
        res: bool = await self._session.request(
            Route("POST", "/api/admin/roles/unassign"), auth=True, json=body
        )
        return res

    async def show(self, role_id: str | None = None) -> Role:
        role_id = self._role_id or role_id
        if role_id is None:
            raise ValueError("role_idは必須です")
        res: IRole = await self._session.request(
            Route("POST", "/api/admin/roles/show"),
            json={"roleId": role_id},
            auth=True,
            lower=True,
        )
        return Role(res, client=self._client)

    async def get_users(
        self,
        role_id: str | None = None,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 100,
        get_all: bool = False,
    ) -> AsyncGenerator[RoleUser, None]:
        role_id = self._role_id or role_id
        if role_id is None:
            raise ValueError("role_idは必須です")

        if limit > 100:
            raise ValueError("limitは100以下である必要があります")

        if get_all:
            limit = 100

        body = {"limit": limit, "sinceId": since_id, "untilId": until_id, "roleId": role_id}

        pagination = Pagination[IRoleUser](
            self._session, Route("POST", "/api/admin/roles/users"), json=body
        )

        while True:
            raw_role_users = await pagination.next()
            for role_user in raw_role_users:
                yield RoleUser(role_user, client=self._client)

            if get_all is False or pagination.is_final:
                break


class AdminRoleActions(AdminRoleModelActions):
    def __init__(self, role_id: str | None = None, *, session: HTTPClient, client: ClientManager):
        super().__init__(role_id=role_id, session=session, client=client)

    async def create(
        self,
        name: str,
        description: str,
        color: str | None = None,
        iconUrl: str | None = None,
        target: Literal["manual", "conditional"] = "manual",
        cond_formula: dict[Any, Any] | None = None,
        is_public: bool = False,
        is_moderator: bool = False,
        is_administrator: bool = False,
        is_explorable: bool = False,
        as_badge: bool = False,
        can_edit_members_by_moderator: bool = False,
        display_order: int = 0,
        policies: dict[Any, Any] | None = None,
    ) -> Role:
        body = {
            "name": name,
            "description": description,
            "color": color,
            "iconUrl": iconUrl,
            "target": target,
            "condFormula": cond_formula or {},
            "isPublic": is_public,
            "isModerator": is_moderator,
            "isAdministrator": is_administrator,
            "isExplorable": is_explorable,
            "asBadge": as_badge,
            "canEditMembersByModerator": can_edit_members_by_moderator,
            "displayOrder": display_order,
            "policies": policies or {},
        }
        res: IRole = await self._session.request(
            Route("POST", "/api/admin/roles/create"),
            auth=True,
            json=body,
            lower=True,
            remove_none=False,
        )
        return Role(res, client=self._client)

    async def get_list(self) -> list[Role]:
        res: list[IRole] = await self._session.request(
            Route("POST", "/api/admin/roles/list"), auth=True, lower=True
        )
        return [Role(i, client=self._client) for i in res]

    async def update_default_policies(self, policies: IPolicies):
        body = {
            "policies": {
                "gtlAvailable": policies.get("gtl_available"),
                "ltlAvailable": policies.get("ltl_available"),
                "canPublicNote": policies.get("can_public_note"),
                "canInvite": policies.get("can_invite"),
                "canManageCustomEmojis": policies.get("can_manage_custom_emojis"),
                "canHideAds": policies.get("can_hide_ads"),
                "driveCapacityMb": policies.get("drive_capacity_mb"),
                "pinLimit": policies.get("pin_limit"),
                "antennaLimit": policies.get("antenna_limit"),
                "wordMuteLimit": policies.get("word_mute_limit"),
                "webhookLimit": policies.get("webhook_limit"),
                "clipLimit": policies.get("clip_limit"),
                "noteEachClipsLimit": policies.get("note_each_clips_limit"),
                "userListLimit": policies.get("user_list_limit"),
                "userEachUserListsLimit": policies.get("user_each_user_lists_limit"),
                "rateLimitFactor": policies.get("rate_limit_factor"),
            }
        }
        res = await self._session.request(
            Route("POST", "/api/admin/roles/update-default-policies"),
            auth=True,
            lower=True,
            json=body,
        )
        return res
